﻿//*****************************************************************************
//
//  By The Dummy Programmer
//  https://www.thedummyprogrammer.com
//
//*****************************************************************************

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Net;
using System.Net.Sockets;
using System.Security.Cryptography.X509Certificates;
using System.Threading.Tasks;
using TDP.BaseServices.Infrastructure.Configuration;
using TDP.BaseServices.Infrastructure.Logging;
using TDP.BaseServices.Infrastructure.Logging.Abstract;
using TDP.BaseServices.Infrastructure.Security;
using TDP.PushLib.Messages;
using TDP.PushLib.Messages.Core;
using TDP.PushLib.Server.Infrastructure;
using TDP.PushLib.Server.Infrastructure.Configuration;

namespace TDP.PushLib.Server
{
    public class PushServer<T>
    {
        #region "Event definitions"
        public event EventHandler InitStarted;
        protected virtual void OnInitStarted(EventArgs e)
        {
            EventHandler handler = InitStarted;
            if (handler != null)
            {
                foreach (EventHandler singleCast in handler.GetInvocationList())
                {
                    ISynchronizeInvoke syncInvoke = singleCast.Target as ISynchronizeInvoke;
                    if ((syncInvoke != null) && (syncInvoke.InvokeRequired))
                        syncInvoke.Invoke(singleCast, new object[] { this, e });
                    else
                        singleCast(this, e);
                }
            }
        }

        public delegate void InitEndedDelegate(object sender, InitEndedEventArgs e);
        public event InitEndedDelegate InitEnded;
        protected virtual void OnInitEnded(InitEndedEventArgs e)
        {
            InitEndedDelegate handler = InitEnded;
            if (handler != null)
            {
                foreach (InitEndedDelegate singleCast in handler.GetInvocationList())
                {
                    ISynchronizeInvoke syncInvoke = singleCast.Target as ISynchronizeInvoke;
                    if ((syncInvoke != null) && (syncInvoke.InvokeRequired))
                        syncInvoke.Invoke(singleCast, new object[] { this, e });
                    else
                        singleCast(this, e);
                }
            }
        }

        public event EventHandler ShutdownStarted;
        protected virtual void OnShutdownStarted(EventArgs e)
        {
            EventHandler handler = ShutdownStarted;
            if (handler != null)
            {
                foreach (EventHandler singleCast in handler.GetInvocationList())
                {
                    ISynchronizeInvoke syncInvoke = singleCast.Target as ISynchronizeInvoke;
                    if ((syncInvoke != null) && (syncInvoke.InvokeRequired))
                        syncInvoke.Invoke(singleCast, new object[] { this, e });
                    else
                        singleCast(this, e);
                }
            }
        }

        public event EventHandler ShutdownEnded;
        protected virtual void OnShutdownEnded(EventArgs e)
        {
            EventHandler handler = ShutdownEnded;
            if (handler != null)
            {
                foreach (EventHandler singleCast in handler.GetInvocationList())
                {
                    ISynchronizeInvoke syncInvoke = singleCast.Target as ISynchronizeInvoke;
                    if ((syncInvoke != null) && (syncInvoke.InvokeRequired))
                        syncInvoke.Invoke(singleCast, new object[] { this, e });
                    else
                        singleCast(this, e);
                }
            }
        }
        #endregion
        private static ILogger _Log = new Logger();
        private ConfigReader _ConfigReader = new ConfigReader();
        private bool _ServerRunning;
        private Socket _SocketServer;
        private List<Client<T>> _ConnectedClients = new List<Client<T>>();
        private object _LockObject = new object();
        private X509Certificate2 _SslCertificate;

        public async Task StartAsync()
        {
            await Task.Run(() =>
            {
                try
                {
                    _Log.Info("PushLibServer -> StartAsync() -> async task started");

                    OnInitStarted(new EventArgs());


                    Stop();

                    _SocketServer = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
                    IPAddress MyIPAddress = IPAddress.Parse(_ConfigReader.Get(ConfigReaderKeys.KeyListenIPAddress, Constants.DefaultListenIPAddress));
                    IPEndPoint MyEndPoint = new IPEndPoint(MyIPAddress, _ConfigReader.Get(ConfigReaderKeys.KeyListenPort, Constants.DefaultListenPort));
                    _SocketServer.Bind(MyEndPoint);
                    _SocketServer.Listen(_ConfigReader.Get(ConfigReaderKeys.KeyListenBacklog, Constants.DefaultListenBacklog));

                    if (_ConfigReader.Get(ConfigReaderKeys.KeyUseSsl, Constants.DefaultUseSSL))
                    {
                        if (_ConfigReader.Get(ConfigReaderKeys.KeyCertificate, Constants.DefaultCertificate) == string.Empty)
                        {
                            _Log.Info("PushLibServer -> StartAsync() -> Missing SSL certificate configuration");
                            return;
                        }

                        _Log.Info($"PushLibServer -> StartAsync() -> Loading certificate {_ConfigReader.Get(ConfigReaderKeys.KeyCertificate, Constants.DefaultCertificate)}");
                        _SslCertificate = new X509Certificate2(_ConfigReader.Get(ConfigReaderKeys.KeyCertificate, Constants.DefaultCertificate), _ConfigReader.Get(ConfigReaderKeys.KeyCertificatePassword, Constants.DefaultCertificatePassword)); 
                    }

                    _ServerRunning = true;
                }
                catch (Exception ex)
                {
                    _Log.Error("PushLibServer -> StartAsync() -> Error during initialization", ex);
                    OnInitEnded(new InitEndedEventArgs(new Result(-1, false, "Initialization failed")));
                    return;
                }

                Task TaskWaitForClient = WaitForClientsAsync();
                    
                OnInitEnded(new InitEndedEventArgs(new Result(0, true, string.Empty)));

                _Log.Info("PushLibServer -> StartAsync() -> async task ending");
            });
        }

        public void Stop()
        {
            try
            {
                if (_SocketServer != null)
                {
                    if (_SocketServer.Connected)
                        _SocketServer.Disconnect(true);

                    _SocketServer.Dispose();
                }

                lock (_LockObject)
                {
                    foreach (Client<T> client in _ConnectedClients)
                    {
                        client.Disconnect();
                    }

                    _ConnectedClients.Clear();
                }
            }
            catch (Exception ex)
            {
                _Log.Error("PushLibServer -> Stop() -> Error stopping the server", ex);
            }

            _ServerRunning = false;
        }

        public async void StopAsync()
        {
            await Task.Run(() =>
            {
                try
                {
                    _Log.Info("PushLibServer -> StopAsync() -> async task started");

                    OnShutdownStarted(new EventArgs());

                    Stop();

                    OnShutdownEnded(new EventArgs());

                    _Log.Info("PushLibServer -> StopAsync() -> async task ending");
                }
                catch (Exception ex)
                {
                    _Log.Error("PushLibServer -> StopAsync() -> Error during stoppage of the server", ex);
                }
            });
        }

        public void SendPushMessage(T clientID, string message)
        {
            //TODO: Need to sync here?
            PushMessage PushMessage = new PushMessage() { Message = message };
            foreach (Client<T> client in _ConnectedClients.Where(client => client.ID.Equals(clientID)))
            {
                client.SendPushMessage(PushMessage);
            }
        }

        private bool ExecRegister(Client<T> client, RegisterClient<T> message)
        {
            RegisterClientResponse Response = new RegisterClientResponse();
            
            bool ValidateAuthToken = _ConfigReader.Get(ConfigReaderKeys.KeyValidateAuthenticationToken, false);
            if (ValidateAuthToken)
            {
                string XmlRSAKey = _ConfigReader.Get(ConfigReaderKeys.KeyValidateAuthenticationToken, string.Empty);
                AsymmetricCryptographyKey Key = new AsymmetricCryptographyKey(XmlRSAKey);
                int TokenDuration = _ConfigReader.Get(ConfigReaderKeys.KeyAuthTokenDurationMinutes, Constants.AuthTokenDurationMinutes);

                if (!message.AuthToken.IsValid(Key, TokenDuration))
                {
                    Response.Success = false;
                    Response.ErrorCode = -1000;
                    Response.Message = "Invalid authentication token.";

                    client.SendObject(Response);
                    return false;
                }
            }
            
            client.ID = message.AuthToken.ClientID;

            Response.Success = true;
            client.SendObject(Response);

            return true;
        }

        private async Task ManageClientAsync(Socket clientSocket)
        {
            await Task.Run(() =>
            {
                _Log.Info("PushLibServer -> ManageClientAsync() -> async task started");

                using (Client<T> ThisClient = new Client<T>(clientSocket, _SslCertificate))
                {
                    try
                    {
                        object Message = ThisClient.ReceiveObject();

                        bool Result = false;
                        if (Message is RegisterClient<T>)
                        {
                            Result = ExecRegister(ThisClient, (RegisterClient<T>)Message);
                        }
                        else
                        {
                            // Couldn't build the RegisterClient message, exit
                            _Log.Error("PushLibServer -> ManageClientAsync() -> Couldn't build the RegisterClient message, exit from ManageClientAsync");
                            return;
                        }

                        // Enter the push notification loop
                        if (Result)
                        {
                            lock (_LockObject)
                            {
                                _ConnectedClients.Add(ThisClient);
                            }
                            
                            bool RunClient = true;
                            while (RunClient)
                            {
                                RunClient = ThisClient.ManagePushMessages();
                            }
                        }
                    }
                    catch (Exception ex)
                    {
                        // Track the error and remove the client
                        _Log.Error("PushLibServer -> ManageClientAsync() -> An error occurred, removing the client from connection list", ex);

                        lock (_LockObject)
                        {
                            _ConnectedClients.Remove(ThisClient);
                        }
                    }
                }
            });
        }

        private async Task WaitForClientsAsync()
        {
            await Task.Run(() =>
            {
                _Log.Info("PushLibServer -> WaitForClientsAsync() -> async task started");

                while (_ServerRunning)
                { 
                    _Log.Info("PushLibServer -> WaitForClientsAsync() -> Waiting for connections");

                    Socket ClientSocket = null;
                    try
                    {
                        ClientSocket = _SocketServer.Accept();
                    }
                    catch (Exception ex)
                    {
                        _Log.Error("PushLibServer -> WaitForClientsAsync() -> Error while waiting for connections, the server will be stopped", ex);
                        Stop();
                        return;
                    }

                    _Log.Info("PushLibServer -> WaitForClientsAsync() -> Connection accepted");

                    Task TaskManageClient = ManageClientAsync(ClientSocket);
                }
            });
        }
    }
}
